@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import hashlib
 import logging
 import stat
 import threading
@@ -164,7 +165,7 @@ def ssh_stream(
         while chan.recv_stderr_ready():
             data = chan.recv_stderr(chunk_size)
             if data:
-                yield ("stdin", data.decode(encoding, errors="replace"))
+                yield ("stderr", data.decode(encoding, errors="replace"))
         if (
             chan.exit_status_ready()
             and not chan.recv_ready()
@@ -189,7 +190,7 @@ def instance_exec(
             match msg:
                 case ("stdout", txt):
                     on_stdout(txt)
-                case ("stdin", txt):
+                case ("stderr", txt):
                     on_stderr(txt)
                 case ("exit_code", code):
                     return code
@@ -242,21 +243,20 @@ class Snapshot:
             for s in snaps:
                 if invalidate_fn(Snapshot(s)):
                     s.delete()
+        digest = f"{name}-{image_id}-{vcpus}-{memory}-{disk_size}"
         snap = client.snapshots.create(
             image_id=image_id,
             vcpus=vcpus,
             memory=memory,
             disk_size=disk_size,
-            digest=name,
+            digest=digest,
             metadata={"name": name},
         )
         return cls(snap)
 
     @classmethod
     def from_snapshot_id(cls, snapshot_id: str) -> "Snapshot":
-        logger.info(
-            "🔍 Snapshot.from_snapshot_id()", extra={"snapshot_id": snapshot_id}
-        )
+        logger.info("🔍 Snapshot.from_snapshot_id()", extra={"snapshot_id": snapshot_id})
         snap = client.snapshots.get(snapshot_id)
         return cls(snap)
 
@@ -312,7 +312,16 @@ class Snapshot:
             yield inst
 
     def key_to_digest(self, key: str) -> str:
-        return (self.snapshot.digest or "") + self.snapshot.id + key
+        """
+        Computes a digest hash based on the parent snapshot and operation key.
+        Follows the same pattern as api.py's compute_chain_hash method.
+        """
+        parent_content = (self.snapshot.digest or "") + self.snapshot.id
+        hasher = hashlib.sha256()
+        hasher.update(parent_content.encode("utf-8"))
+        hasher.update(b"\n")
+        hasher.update(key.encode("utf-8"))
+        return hasher.hexdigest()
 
     def apply(
         self,
@@ -324,6 +333,7 @@ class Snapshot:
             None,
         ] = None,
         invalidate: InvalidateFn | bool = False,
+        debug: bool = False,
     ):
         invalidate_fn = (
             invalidate
@@ -345,27 +355,53 @@ class Snapshot:
                 return Snapshot(snaps[0])
 
         if start_fn is None:
-            context_manager = self.start()
+            context_manager = self.start(ttl_seconds=24 * 60 * 60, ttl_action="stop")
         elif callable(start_fn):
-            context_manager = start_fn()
+            context_manager = start_fn(ttl_seconds=24 * 60 * 60, ttl_action="stop")
         else:
             context_manager = start_fn
 
-        with context_manager as inst:
-            res = func(inst)
-            inst = inst if res is None else res
-            return Snapshot(
-                inst.snapshot(digest=self.key_to_digest(key) if key else None)
-            )
+        if debug:
+            inst = context_manager.__enter__()
+            try:
+                res = func(inst)
+                inst = inst if res is None else res
+
+                new_snapshot = Snapshot(
+                    inst.snapshot(digest=self.key_to_digest(key) if key else None)
+                )
+
+                logger.info(
+                    f"Debug mode - Command succeeded, stopping instance {inst.id}"
+                )
+                context_manager.__exit__(None, None, None)
+
+                return new_snapshot
+
+            except Exception as exc:
+                logger.warning(
+                    f"Debug mode - Command failed, leaving instance {inst.id} running for debugging "
+                    f"(24-hour TTL active): {exc}"
+                )
+                raise
+        else:
+            with context_manager as inst:
+                res = func(inst)
+                inst = inst if res is None else res
+                return Snapshot(
+                    inst.snapshot(digest=self.key_to_digest(key) if key else None)
+                )
 
     # -------------- run with stream between CMD/RET -------------- #
-    def run(self, command: str, invalidate: InvalidateFn | bool = False):
-        logger.info("🚀 Snapshot.run()", extra={"command": command})
+    def run(
+        self, command: str, debug: bool = False, invalidate: InvalidateFn | bool = False
+    ):
+        logger.info("🚀 Snapshot.run()", extra={"command": command, "debug_mode": debug})
 
         def execute(instance):
             logger.info(
                 "🖥  Snapshot.run() - Starting command execution",
-                extra={"command": command},
+                extra={"command": command, "debug_mode": debug},
             )
 
             buf = deque()
@@ -379,7 +415,7 @@ class Snapshot:
             exit_code = instance_exec(instance, command, _out, _err)
             logger.info(
                 "🖥  Snapshot.run() - Command completed",
-                extra={"command": command, "exit_code": exit_code},
+                extra={"command": command, "exit_code": exit_code, "debug_mode": debug},
             )
 
             if exit_code != 0:
@@ -389,7 +425,7 @@ class Snapshot:
                     f"Command execution failed: {command} exit={exit_code} recent_output={recent_output}"
                 )
 
-        return self.apply(execute, key=command, invalidate=invalidate)
+        return self.apply(execute, key=command, invalidate=invalidate, debug=debug)
 
     def copy_(self, src: str, dest: str, invalidate: InvalidateFn | bool = False):
         """
@@ -579,9 +615,7 @@ class Snapshot:
             return all_ok
 
         def run_verification(instance):
-            logger.info(
-                "🔍 Starting verification", extra={"instructions": instructions}
-            )
+            logger.info("🔍 Starting verification", extra={"instructions": instructions})
             success = verifier(instance)
             if not success:
                 raise Exception("Verification failed.")
